using System;
using System.IO;
using System.Net;
using System.Text;
using System.Threading;
using System.Net.Sockets;
using System.Net.Security;
using System.Collections.Generic;
using System.Security.Cryptography.X509Certificates;

namespace Connection
{
	public class SslStreamConnection
	{
#region Variables
		private TcpClient		tcpClient;
		private TcpListener		tcpListener;
        private SslStream		sslStream;
        private StreamReader 	sr;
        private StreamWriter 	sw;
        private Thread			readingThread;
#endregion

// --------------------------------------------------------------------------- //

#region Properties
		private string host;
		public string Host
		{
			get
			{
				return host;
			}
			set
			{
				host = value;
			}
		}
	
// ------------------------------------------------------------------ //

		private UInt16 port;
		public UInt16 Port
		{
			get
			{
				return port;
			}
			set
			{
				port = value;
			}
		}

// ------------------------------------------------------------------ //

		private int timeout = 15000;
		public int Timeout
		{
			get
			{
				return timeout;
			}
			set
			{
				timeout = value;
			}
		}

// ------------------------------------------------------------------ //

		private bool isConnected; // false by default
		public bool IsConnected
		{
			get
			{
				return isConnected;
			}
			set
			{
				isConnected = value;
			}
		}

// ------------------------------------------------------------------ //

		private bool isSSL; // false by default
		public bool IsSSL
		{
			get
			{
				return isSSL;
			}
			set
			{
				isSSL = value;
			}
		}

// ------------------------------------------------------------------ //

		private Exception lastException;
		public Exception LastException
		{
			get
			{
				return lastException;
			}
			set
			{
				lastException = value;
			}
		}

// ------------------------------------------------------------------ //

		private IPAddress listeningIp = IPAddress.Any;
		public IPAddress ListeningIp
		{
			get
			{
				return listeningIp;
			}
			set
			{
				listeningIp = value;
			}
		}

// ------------------------------------------------------------------ //

		private UInt16 listeningPort; // 0 by default
		public UInt16 ListeningPort
		{
			get
			{
				return listeningPort;
			}
			set
			{
				listeningPort = value;
			}
		}

// ------------------------------------------------------------------ //

		private bool manualReading; // false by default
		public bool ManualReading
		{
			get
			{
				return manualReading;
			}
			set
			{
				manualReading = value;
			}
		}

// ------------------------------------------------------------------ //

		private IPEndPoint localEndPoint;
		public IPEndPoint LocalEndPoint
		{
			get
			{
				return localEndPoint;
			}
			set
			{
				localEndPoint = value;
			}
		}

// ------------------------------------------------------------------ //

		private IPEndPoint remoteEndPoint;
		public IPEndPoint RemoteEndPoint
		{
			get
			{
				return remoteEndPoint;
			}
			set
			{
				remoteEndPoint = value;
			}
		}

// ------------------------------------------------------------------ //
#endregion		

// --------------------------------------------------------------------------- //

#region Constructor
		public SslStreamConnection (string host, UInt16 port, bool isSsl)
        {
            this.host	= host;
            this.port	= port;
			this.isSSL	= isSsl;
		}

// --------------------------------------------------------------------------- //

		public SslStreamConnection (IPEndPoint ipe, bool isSsl):this (ipe.Address.ToString (), (UInt16)ipe.Port, isSsl)
        {

		}

// --------------------------------------------------------------------------- //
#endregion

// --------------------------------------------------------------------------- //

#region Methods
        public bool Connect ()
        {
            isConnected = false;

            try
            {
                tcpClient = new TcpClient ();
                tcpClient.Connect (host, port);
                tcpClient.ReceiveTimeout	= timeout;
                tcpClient.SendTimeout		= timeout;

                if (isSSL == true)
                {
                    sslStream = new SslStream (tcpClient.GetStream (), false, CertificateValidation);
                    sslStream.ReadTimeout = timeout;
                    sslStream.WriteTimeout = timeout;

                    // I don't care about the certificate
                    sslStream.AuthenticateAsClient ("", null, System.Security.Authentication.SslProtocols.Ssl2 |
                                                                    System.Security.Authentication.SslProtocols.Ssl3 |
                                                                    System.Security.Authentication.SslProtocols.Tls, false);

                    sw = new StreamWriter (sslStream);
                    sr = new StreamReader (sslStream);
                }
                else
                {
                    sw = new StreamWriter (tcpClient.GetStream ());
                    sr = new StreamReader (tcpClient.GetStream ());
                }


                // without this, the connection will block on SendLine
                sw.NewLine = "\r\n";
                sw.AutoFlush = true;
                
                isConnected = true;

				localEndPoint	= (IPEndPoint)tcpClient.Client.LocalEndPoint;
				remoteEndPoint	= (IPEndPoint)tcpClient.Client.RemoteEndPoint;

                OnConnect ();

				if (manualReading == false)
					StartReading ();
            }
            catch (Exception e)
            {
                lastException = e;
            }

            return isConnected;
        }

// ------------------------------------------------------------------ //

		private bool CertificateValidation (object sender, X509Certificate certificate, X509Chain chain, SslPolicyErrors sslPolicyErrors)
        {
            return true; // I don't care about the certificate, so it will always return true
        }

// ------------------------------------------------------------------ //

		public void Disconnect ()
        {
            try
			{
				sslStream.Close ();
				sw.Close ();
				sr.Close ();
			}
			catch (Exception ex)
			{
				lastException = ex;
			}
            
			OnDisconnect ();

            isConnected = false;
        }

// ----------------------------------------------------------------- //

		private void StartReading ()
        {
            readingThread = new Thread (new ThreadStart (ReadThread));
			readingThread.IsBackground = true;
            readingThread.Start ();
        }

// ----------------------------------------------------------------- //

        private void ReadThread ()
        {
            string line = "";

            while ((isConnected == true) || (isSSL == false))
            {
                try
                {
                    line = sr.ReadLine ();

                    OnReadLine (line + "\r\n"); // the line is without the new line sequence
                }
				catch (IOException ioe)
                {
					lastException = ioe;

					line = null;
                    isConnected = false;
                }
				catch (Exception e)
                {
					lastException = e;
                    line = null;
                    isConnected = false;
                }
            }

			if (isSSL == true)
				isConnected = true;
			else
				OnDisconnect ();
        }

// ------------------------------------------------------------------ //

		public string ReadLine ()
		{
			string line = "";

			if (isConnected == false)
				return line;

			try
            {
                line = sr.ReadLine ();
            }
			catch (IOException ioe)
            {
				lastException = ioe;

				line = "";
                isConnected = false;
            }
			catch (Exception e)
            {
				lastException = e;
                line = "";
                isConnected = false;
            }

			return line;
		}

// ------------------------------------------------------------------ //

		public int Read (char[] buffer, int length)
		{
			if (isConnected == false)
				return 0;

			int bytesRead = 0;

			try
            {
                bytesRead = sr.Read (buffer, 0, length);
            }
			catch (IOException ioe)
            {
				lastException = ioe;

				bytesRead = 0;
                isConnected = false;
            }
			catch (Exception e)
            {
				lastException = e;
                bytesRead = 0;
                isConnected = false;
            }

			return bytesRead;
		}

// ------------------------------------------------------------------ //

		public bool SendLine (string buffer)
        {
            try
            {
                sw.WriteLine (buffer);
				OnSendLine (buffer + "\r\n");

				return true;
            }
            catch (SocketException se)
            {
				lastException = se;
            }
            catch (Exception e)
            {
				lastException = e;
            }

			OnDisconnect ();

			return false;
		}

// ------------------------------------------------------------------ //

		public void EnableSecureLayer ()
		{
			sslStream = new SslStream (tcpClient.GetStream (), false, CertificateValidation);
			sslStream.ReadTimeout = timeout;
			sslStream.WriteTimeout = timeout;

			// I don't care about the certificate
			sslStream.AuthenticateAsClient ("hello", null,	System.Security.Authentication.SslProtocols.Ssl2 |
														System.Security.Authentication.SslProtocols.Ssl3 |
														System.Security.Authentication.SslProtocols.Tls, false);

			sw = new StreamWriter (sslStream);
			sr = new StreamReader (sslStream);

			sw.AutoFlush = true;

			isSSL = true;

			//readingThread = new Thread (new ThreadStart (SslReadThread));
			//readingThread.IsBackground = true;
			//readingThread.Start ();
		}

// ------------------------------------------------------------------ //

		public IPEndPoint Listen (IPAddress ip, UInt16 port)
		{
			IPEndPoint ipe = new IPEndPoint (ip, port);

			return Listen (ipe);
		}

// ------------------------------------------------------------------ //

		public IPEndPoint Listen (IPEndPoint ipe)
		{
			try
			{
				tcpListener = new TcpListener (ipe);
				tcpListener.Start (5);
			}
			catch
			{
				return null;
			}

			return (IPEndPoint)tcpListener.LocalEndpoint;
		}

// ------------------------------------------------------------------ //

		public bool AcceptClient ()
		{
			tcpClient = tcpListener.AcceptTcpClient ();

			sw = new StreamWriter (tcpClient.GetStream ());
            sr = new StreamReader (tcpClient.GetStream ());
            
            // without this, the connection will block on SendLine
            sw.NewLine = "\r\n";
            sw.AutoFlush = true;

			isConnected = true;

			return true;
		}

// ------------------------------------------------------------------ //

		public void StopListening ()
		{
			tcpListener.Stop ();
		}

// ------------------------------------------------------------------ //

#endregion

// --------------------------------------------------------------------------- //

#region Events
	#region ConnectEventHandler
		public delegate void ConnectedEvent (Object sender, ConnectedEventArgs e);
		public event ConnectedEvent	Connected;

		protected virtual void OnConnect ()
		{
			if (Connected != null)
			{
				ConnectedEventArgs e = new ConnectedEventArgs ();
				e.Host = host;
				e.Port = port;

				Connected (this, e);
			}
		}

		public class ConnectedEventArgs : System.EventArgs
		{
			private string host;
			public string Host
			{
				get
				{
					return host;
				}
				set
				{
					host = value;
				}
			}

			private int port;
			public int Port
			{
				get
				{
					return port;
				}
				set
				{
					port = value;
				}
			}
		}
	#endregion

	#region DisconnectEventHandler
		public delegate void DisconnectedEvent (Object sender, DisconnectedEventArgs e);
		public event DisconnectedEvent Disconnected;

		protected virtual void OnDisconnect ()
		{
			if (Disconnected != null)
			{
				DisconnectedEventArgs e = new DisconnectedEventArgs ();
				e.Host = host;
				e.Port = port;
				e.Error = lastException;

				Disconnected (this, e);
			}
		}

		public class DisconnectedEventArgs : ConnectedEventArgs
		{
			private Exception error;
			public Exception Error
			{
				get
				{
					return error;
				}
				set
				{
					error = value;
				}
			}
		}
	#endregion

	#region ReadLineEventHandler
		public delegate void LineReadEvent (Object sender, LineReadEventArgs e);
		public event LineReadEvent LineRead;

		protected virtual void OnReadLine (string line)
		{
			if (LineRead != null)
			{
				LineReadEventArgs e = new LineReadEventArgs ();
				e.Line = line;
				e.Host = host;
				e.Port = port;

				LineRead (this, e);
			}
		}

		public class LineReadEventArgs : ConnectedEventArgs
		{
			private string line;
			public string Line
			{
				get
				{
					return line;
				}
				set
				{
					line = value;
				}
			}
		}
	#endregion

	#region SendLineEventHandler
		public delegate void LineSentEvent (Object sender, LineSentEventArgs e);
		public event LineSentEvent LineSent;

		protected virtual void OnSendLine (string line)
		{
			if (LineSent != null)
			{
				LineSentEventArgs e = new LineSentEventArgs ();
				e.Line = line;
				e.Host = host;
				e.Port = port;

				LineSent (this, e);
			}
		}

		public class LineSentEventArgs : ConnectedEventArgs
		{
			private string line;
			public string Line
			{
				get
				{
					return line;
				}
				set
				{
					line = value;
				}
			}
		}
	#endregion
#endregion

// --------------------------------------------------------------------------- //
	}
}
